"""pygoto - Module to control GOTO scopes"""

import math
import sys

import serial

class PyGOTO(object):
    """PyGOTO - Contains connection and methods to control GOTO scopes"""

    right_ascension_id, declination_id = range(0, 2)

    TRACK_OFF, TRACK_ALT_AZ, TRACK_EQ_N, TRACK_EQ_S = range(0, 4)

    def __init__(self, dev):
        self.device = serial.Serial(dev, timeout=5)

    def echo(self, val):
        """echo - Scope returns the value it's sent

        Arguments:
            val: value for scope to echo

        Return:
            value from scope
        """

        #print("K{}".format(val))
        #import code; code.interact(local=dict(globals(), **locals()))
        self.__write_to_scope("K{}".format(val))
        return self.__read_from_scope()

    def get_ra_dec(self):
        """get_ra_dec - Get current Right Ascension and Declination coordinates from scope

        Return:
            array containing float values of [Right Ascension, Declination]
        """

        return self.__recv_coord("E")

    def get_precise_ra_dec(self):
        """get_precise_ra_dec - Get current precise Right Ascentions and Declination coordinates
            from scope

        This method returns a more precise coordinate set than get_ra_dec()

        Return:
            array containing more precise float values of [Right Ascension, Declination]
        """

        return self.__recv_coord("e", True)

    def get_alt_az(self):
        """get_alt_az - Get current Altitude and Azimuth coordinates from scope

        Return:
            array containing float values of [Altitude, Azimuth]
        """

        return self.__recv_coord("Z")

    def get_precise_alt_az(self):
        """get_precise_alt_az - Get current precise Altitude and Azimuth coordinates from scope

        This method returns a more precise coordinate set than get_alt_az

        Return:
            array containing more precise float values of [Altitude, Azimuth]
        """

        return self.__recv_coord("z", True)

    def goto_ra_dec(self, right_ascension, declination):
        """goto_ra_dec - Point scope to new coordinates specified

        Arguments:
            right_ascension: float containing target Right Ascension
            declination: float containing target Declination
        """

        self.__send_coord("R", right_ascension, declination)

    def goto_precise_ra_dec(self, right_ascension, declination):
        """goto_precise_ra_dec - Point scope to a more precise set of coordinates

        This method accepts a more precise set of coordinates than goto_ra_dec()

        Arguments:
            right_ascension: float containing target Right Ascension
            declination: float containing target Declination
        """

        self.__send_coord("r", right_ascension, declination, True)

    def goto_azm_alt(self, azm, alt):
        """goto_azm_alt - Point scope to an object at the specified Azimuth/Altitude coordinaes

        Arguments:
            azm: float containing target Altitude
            alt: float containing target Azimuth
        """

        self.__send_coord("B", azm, alt)

    def goto_precise_azm_alt(self, azm, alt):
        """goto_precise_azm_alt - Point scope to an object at the specified precise
            Azimuth/Altitude coordinates

        This method accepts a more precise set of coordinates than goto_ra_dec()

        Arguments:
            azm: float containing target Altitude
            alt: float containing target Azimuth
        """

        self.__send_coord("b", azm, alt, True)

    def sync_ra_dec(self, right_ascension, declination):
        """sync_ra_dec - Synchronize Right Ascension/Declination

        Tell scope that it's current coordinates are those specified
        """

        self.__send_coord("S", right_ascension, declination)

    def sync_precise_ra_dec(self, right_ascension, declination):
        """sync_precise_ra_dec - Sychronize Right Ascension/Declination (precise)


        Tell scope that it's current coordinates are those specified.
        This method accepts a more precise set of coordinates than sync_ra_dec()
        """

        self.__send_coord("s", right_ascension, declination, True)

    def get_tracking_mode(self): #pylint: disable=R0201
        """get_tracking_mode - Get current tracking mode

        Note: Incomplete

        Returns:
            current tracking mode:
                TODO: Better elaborate these returns (specify constants from above?)
                0: No tracking
                1: Tracking Altitude/Azimuth
                2: Tracking Equitorial - Northern hemisphere
                3: Tracking Equitorial - Southern hemisphere
        """
        return False

    #    self.__write_to_scope("t")
    #    return ord(self.__read_from_scope())

    def set_tracking_mode(self, mode):
        """set_tracking_mode - Set scope's tracking mode

        Arguments:
            mode: New tracking mode:
                TODO: Better elaborate these arguments (specify constants from above?)
                0: No tracking
                1: Tracking Altitude/Azimuth
                2: Tracking Equitorial - Northern hemisphere
                3: Tracking Equitorial - Southern hemisphere

        Return:
        """

        self.__write_to_scope("T{}".format(chr(mode)))
        self.__read_from_scope()

    def variable_rate_slew(self, rate, axis=right_ascension_id, positive=True):
        """variable_rate_slew - TODO

        Arguments:
            rate: float containing slew rate in arseconds/second
            axis: right_ascension_id (for Right Ascension or Azimuth), declination_id
                (for Declination or Altitude)
            positive: True for positive direction; False for negative direction

        Return:
        """

        self.__write_to_scope("P{}{}{}{}{}{}".format(
            chr(3),
            chr(16) if axis == self.right_ascension_id else chr(17),
            chr(6) if positive else chr(7),
            chr(int((rate * 4) // 256)),
            chr(int((rate * 4) % 256)),
            chr(0) * 2))
        self.__read_from_scope()

    def fixed_rate_slew(self, rate, axis=right_ascension_id, positive=True):
        """fixed_rate_slew

        Arguments:
            rate: slew rate (-9..9)
            axis: RA (for Right Ascension or Azimuth), DEC (for Declination or Altitude)
            positive: True for positive direction; False for negative direction

        Return:
        """

        self.__write_to_scope("P{}{}{}{}{}".format(
            chr(2),
            chr(16) if axis == self.right_ascension_id else chr(17),
            chr(36) if positive else chr(37),
            chr(int(rate)),
            chr(0) * 3))
        self.__read_from_scope()

    def get_location(self): #pylint: disable=R0201
        """get_location - Get current location of scope (GPS?)

        Note: Incomplete

        Return:
        """

        return False
    #    self.__write_to_scope("w")
    #    return map(ord, list(self.__read_from_scope))

    def set_location(self, lat_d, lat_m, lat_s, hemi_ns, lon_d, lon_m, lon_s, hemi_ew): #pylint: disable=R0913
        """set_location - Set current location of scope (GPS?)

        Arguments:
        """

        self.__write_to_scope("W{}{}{}{}{}{}{}{}".format(
            chr(lat_d),
            chr(lat_m),
            chr(lat_s),
            chr(hemi_ns),
            chr(lon_d),
            chr(lon_m),
            chr(lon_s),
            chr(hemi_ew)))
        self.__read_from_scope()

    def get_time(self): #pylint: disable=R0201
        """get_time - Get current time from scope

        Note: Incomplete

        Return:
        """

        return False
    #    self.__write_to_scope("h")
    #    return map(ord, list(self.__read_from_scope))

    def set_time(self, hour, minute, second, month, day, year, gmt_offset, dst=False): #pylint: disable=R0913
        """set_time - Set current time on scope

        Arguments:

        Return:
        """
        self.__write_to_scope("H{}{}{}{}{}{}{}{}".format(
            chr(hour),
            chr(minute),
            chr(second),
            chr(month),
            chr(day),
            chr(year),
            chr(gmt_offset) if gmt_offset >= 0 else chr(256+gmt_offset),
            chr(1) if dst else chr(0)))
        self.__read_from_scope()


    def __recv_coord(self, cmd, precise=False):
        self.__write_to_scope(cmd.lower() if precise else cmd.upper())
        hex_coord1, hex_coord2 = self.__read_from_scope().split(',')
        coord1 = self.__hex_to_float("{}{}".format(
            hex_coord1[:6] if precise else hex_coord1,
            "" if precise else "00"))
        coord2 = self.__hex_to_float("{}{}".format(
            hex_coord2[:6] if precise else hex_coord2,
            "" if precise else "00"))
        return [coord1, coord2]

    def __send_coord(self, cmd, coord1, coord2, precise=False):
        hex_coord1 = self.__float_to_hex(coord1)
        hex_coord2 = self.__float_to_hex(coord2)

        self.__write_to_scope("{}{}{},{}{}".format(
            cmd.lower() if precise else cmd.upper(),
            hex_coord1 if precise else hex_coord1[:4],
            "00" if precise else "",
            hex_coord2 if precise else hex_coord2[:4],
            "00" if precise else ""))
        self.__read_from_scope()

    def __read_from_scope(self):
        result = []
        for character in self.device.read().decode('utf-8'):
            #print(character)

            if character == '#':
                break
            else:
                result.append(character)

        return ''.join(result)

    def __write_to_scope(self, command):
        if sys.version_info[0] > 2:
            command = command.encode()
        self.device.write(command)

    @classmethod
    def __float_to_hex(cls, val):
        #import code; code.interact(local=dict(globals(), **locals()))
        return "{:06X}".format(int(math.floor(val * 16777216.0 / 360.0)))

    @classmethod
    def __hex_to_float(cls, val):
        return int(val, 16) / 16777216.0
